@@ -22,12 +22,17 @@
 #include "Rversion.h"
 
 #include "Rinternals.h"
+#include "R_ext/GraphicsEngine.h"
+
 #if R_VERSION < R_Version(2,7,0)
 # include "Rgraphics.h"
 # include "Rdevices.h"
 # include "R_ext/GraphicsDevice.h"
+typedef GEDevDesc* pGEDevDesc;
+typedef NewDevDesc* pDevDesc;
+typedef R_GE_gcontext* pGEcontext;
 #endif
-#include "R_ext/GraphicsEngine.h"
+
 #include  <ctype.h> /* for islower */
 
 #ifndef BEGIN_SUSPEND_INTERRUPTS
@@ -496,49 +501,49 @@ static char* toolTip2Header[] = {
 
 /* Device driver actions */
 
-static void   SVG_Activate(NewDevDesc *);
+static void   SVG_Activate(pDevDesc dd);
 static void   SVG_Circle(double x, double y, double r,
-                         R_GE_gcontext *gc,
-                         NewDevDesc *dd);
-static void   SVG_Clip(double, double, double, double, NewDevDesc*);
-static void   SVG_Close(NewDevDesc*);
-static void   SVG_Deactivate(NewDevDesc *);
+                         const pGEcontext gc,
+                         pDevDesc dd);
+static void   SVG_Clip(double, double, double, double, pDevDesc);
+static void   SVG_Close(pDevDesc);
+static void   SVG_Deactivate(pDevDesc );
 static void   SVG_Line(double x1, double y1, double x2, double y2,
-                       R_GE_gcontext *gc,
-                       NewDevDesc *dd);
-static Rboolean SVG_Locator(double*, double*, NewDevDesc*);
-static void   SVG_Mode(int, NewDevDesc*);
-static void   SVG_NewPage(R_GE_gcontext *gc, NewDevDesc *dd);
-static Rboolean SVG_Open(NewDevDesc*, SVGDesc*);
+                       const pGEcontext gc,
+                       pDevDesc dd);
+static Rboolean SVG_Locator(double*, double*, pDevDesc);
+static void   SVG_Mode(int, pDevDesc);
+static void   SVG_NewPage(const pGEcontext gc, pDevDesc dd);
+static Rboolean SVG_Open(pDevDesc, SVGDesc*);
 static void   SVG_Polygon(int n, double *x, double *y,
-                          R_GE_gcontext *gc,
-                          NewDevDesc *dd);
+                          const pGEcontext gc,
+                          pDevDesc dd);
 static void   SVG_Polyline(int n, double *x, double *y,
-                           R_GE_gcontext *gc,
-                           NewDevDesc *dd);
+                           const pGEcontext gc,
+                           pDevDesc dd);
 static void   SVG_Rect(double x0, double y0, double x1, double y1,
-                       R_GE_gcontext *gc,
-                       NewDevDesc *dd);
+                       const pGEcontext gc,
+                       pDevDesc dd);
 static void   SVG_Size(double *left, double *right,
                        double *bottom, double *top,
-                       NewDevDesc *dd);
+                       pDevDesc dd);
 
 /*
 static void   SVG_Resize(double *left, double *right,
                          double *bottom, double *top,
-                         NewDevDesc *dd);
+                         pDevDesc dd);
 */
 static double SVG_StrWidth(const char *str,
-                           R_GE_gcontext *gc,
-                           NewDevDesc *dd);
+                           const pGEcontext gc,
+                           pDevDesc dd);
 static void   SVG_Text(double x, double y, const char *str,
                        double rot, double hadj,
-                       R_GE_gcontext *gc,
-                       NewDevDesc *dd);
+                       const pGEcontext gc,
+                       pDevDesc dd);
 static void   SVG_MetricInfo(int c,
-                             R_GE_gcontext *gc,
+                             const pGEcontext gc,
                              double* ascent, double* descent,
-                             double* width, NewDevDesc *dd);
+                             double* width, pDevDesc dd);
 
 /* Support routines */
 
@@ -561,7 +566,7 @@ char *col2RGBname(unsigned int col)
 /*Thanks Paul*/
 static void SVG_Size(double *left, double *right,
                      double *bottom, double *top,
-                     NewDevDesc *dd)
+                     pDevDesc dd)
 {
     *left = dd->left;
     *right = dd->right;
@@ -571,7 +576,7 @@ static void SVG_Size(double *left, double *right,
 
 void SetSvgShapeContents(char **str)
 {
-    GEDevDesc *dd = GEcurrentDevice();
+    pGEDevDesc dd = GEcurrentDevice();
     SVGDesc *ptd;
     if (dd==0 || dd->dev==0)
 	return;
@@ -590,7 +595,7 @@ void SetSvgShapeContents(char **str)
 
 void SetSvgShapeURL(char **str)
 {
-    GEDevDesc *dd = GEcurrentDevice();
+    pGEDevDesc dd = GEcurrentDevice();
     SVGDesc *ptd;
     if (dd==0 || dd->dev==0)
 	return;
@@ -628,7 +633,7 @@ void SetSvgShapeURLTarget(char **str)
 
 void GetSvgToolTipMode(int *mode)
 {
-    GEDevDesc *dd = GEcurrentDevice();
+    pGEDevDesc dd = GEcurrentDevice();
     SVGDesc *ptd;
     /* this might get called when some other graphics device is being used, */
     /* in which case we want to return mode = -1 */
@@ -644,7 +649,7 @@ void GetSvgToolTipMode(int *mode)
 /*Get Device point from user point*/
 /*void GetSvgDevicePoint(double x, double y)
   {
-  GEDevDesc *dd = GEcurrentDevice();
+  pGEDevDesc dd = GEcurrentDevice();
 
   x = toDeviceX(x,
   GConvert(x, y, GE_NDC, GE_DEVICE, dd);
@@ -663,7 +668,7 @@ void GetSvgToolTipMode(int *mode)
   {
   int i;
 
-  GEDevDesc *dd = GEcurrentDevice();
+  pGEDevDesc dd = GEcurrentDevice();
 
   for(i = 0; i < *n; i++){
   GConvert(x+i, y+i, GE_NDC, DEVICE, dd);
@@ -680,7 +685,7 @@ void GetSvgToolTipMode(int *mode)
   *h = in2dots(ptd->height);
   }
 */
-static void SetLinetype(int newlty, int newlwd, NewDevDesc *dd, int
+static void SetLinetype(int newlty, int newlwd, pDevDesc dd, int
                         fgcol, int col)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
@@ -786,18 +791,18 @@ static void SetFont(int face, int size, SVGDesc *ptd)
     ptd->fontface = lface;
 }
 
-static void SVG_Activate(NewDevDesc *dd)
+static void SVG_Activate(pDevDesc dd)
 {
 }
 
-static void SVG_Deactivate(NewDevDesc *dd)
+static void SVG_Deactivate(pDevDesc dd)
 {
 }
 
 static void SVG_MetricInfo(int c,
-                           R_GE_gcontext *gc,
+                           const pGEcontext gc,
                            double* ascent, double* descent,
-                           double* width, NewDevDesc *dd)
+                           double* width, pDevDesc dd)
 {
     /* metric information not available => return 0,0,0 */
     *ascent  = 0.0;
@@ -874,7 +879,7 @@ static void SVG_footer(SVGDesc *ptd)
 
 /* Initialize the device */
 
-static Rboolean SVG_Open(NewDevDesc *dd, SVGDesc *ptd)
+static Rboolean SVG_Open(pDevDesc dd, SVGDesc *ptd)
 {
     ptd->fontsize = 0;
     ptd->fontface = 0;
@@ -902,14 +907,14 @@ static Rboolean SVG_Open(NewDevDesc *dd, SVGDesc *ptd)
 /*
 static void SVG_Resize(double *left, double *right,
                        double *bottom, double *top,
-                       NewDevDesc *dd)
+                       pDevDesc dd)
 {
 
 }
 */
 
 static void SVG_Clip(double x0, double x1, double y0, double y1,
-                     NewDevDesc *dd)
+                     pDevDesc dd)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
 
@@ -921,7 +926,7 @@ static void SVG_Clip(double x0, double x1, double y0, double y1,
 
 /* Start a new page */
 
-static void SVG_NewPage(R_GE_gcontext *gc, NewDevDesc *dd)
+static void SVG_NewPage(const pGEcontext gc, pDevDesc dd)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
     int face, size;
@@ -941,7 +946,7 @@ static void SVG_NewPage(R_GE_gcontext *gc, NewDevDesc *dd)
 
 /* Close down the driver */
 
-static void SVG_Close(NewDevDesc *dd)
+static void SVG_Close(pDevDesc dd)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
 
@@ -962,8 +967,8 @@ static void SVG_Close(NewDevDesc *dd)
 
 
 static void SVG_Line(double x1, double y1, double x2, double y2,
-                     R_GE_gcontext *gc,
-                     NewDevDesc *dd)
+                     const pGEcontext gc,
+                     pDevDesc dd)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
 
@@ -978,8 +983,8 @@ static void SVG_Line(double x1, double y1, double x2, double y2,
 }
 
 static void SVG_Polyline(int n, double *x, double *y,
-                         R_GE_gcontext *gc,
-                         NewDevDesc *dd)
+                         const pGEcontext gc,
+                         pDevDesc dd)
 {
     int i;
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
@@ -999,8 +1004,8 @@ static void SVG_Polyline(int n, double *x, double *y,
 /* For the current font in pointsize fontsize */
 
 static double SVG_StrWidth(const char *str,
-                           R_GE_gcontext *gc,
-                           NewDevDesc *dd)
+                           const pGEcontext gc,
+                           pDevDesc dd)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
 
@@ -1019,8 +1024,8 @@ static double SVG_StrWidth(const char *str,
 
 /* Possibly Filled Rectangle */
 static void SVG_Rect(double x0, double y0, double x1, double y1,
-                     R_GE_gcontext *gc,
-                     NewDevDesc *dd)
+                     const pGEcontext gc,
+                     pDevDesc dd)
 {
     double tmp;
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
@@ -1066,8 +1071,8 @@ static void SVG_Rect(double x0, double y0, double x1, double y1,
 }
 
 static void SVG_Circle(double x, double y, double r,
-                       R_GE_gcontext *gc,
-                       NewDevDesc *dd)
+                       const pGEcontext gc,
+                       pDevDesc dd)
 {
     SVGDesc *ptd = (SVGDesc *) dd->deviceSpecific;
 
@@ -1101,8 +1106,8 @@ static void SVG_Circle(double x, double y, double r,
 }
 
 static void SVG_Polygon(int n, double *x, double *y,
-                        R_GE_gcontext *gc,
-                        NewDevDesc *dd)
+                        const pGEcontext gc,
+                        pDevDesc dd)
 {
     int i;
 
@@ -1195,8 +1200,8 @@ static void textext(const char *str, SVGDesc *ptd)
 
 static void SVG_Text(double x, double y, const char *str,
                      double rot, double hadj,
-                     R_GE_gcontext *gc,
-                     NewDevDesc *dd)
+                     const pGEcontext gc,
+                     pDevDesc dd)
 {
     int size;
 
@@ -1233,17 +1238,17 @@ static void SVG_Text(double x, double y, const char *str,
 }
 
 /* Pick */
-static Rboolean SVG_Locator(double *x, double *y, NewDevDesc *dd)
+static Rboolean SVG_Locator(double *x, double *y, pDevDesc dd)
 {
     return FALSE;
 }
 
 /* Set Graphics mode - not needed for PS */
-static void SVG_Mode(int mode, NewDevDesc* dd)
+static void SVG_Mode(int mode, pDevDesc dd)
 {
 }
 
-Rboolean SVGDeviceDriver(NewDevDesc *dd, char *filename, char *bg, char *fg,
+Rboolean SVGDeviceDriver(pDevDesc dd, char *filename, char *bg, char *fg,
                          double width, double height, Rboolean debug,
                          Rboolean xmlHeader, char *title, int toolTipMode,
 			 int tipFontSize, double tipOpacity,
@@ -1327,11 +1332,7 @@ Rboolean SVGDeviceDriver(NewDevDesc *dd, char *filename, char *bg, char *fg,
     /* We use printer points, i.e. 72.27 dots per inch : */
     dd->ipr[0] = dd->ipr[1] = 1./DOTSperIN;
 
-    dd->canResizePlot = FALSE;
-    dd->canChangeFont = TRUE;
-    dd->canRotateText = TRUE;
-    dd->canResizeText = TRUE;
-    dd->canClip = FALSE; /* theoretically, SVG can clip, but we'd probably need to do a lot of work to use it */
+    dd->canClip = FALSE;
     dd->canHAdj = 0;
     dd->canChangeGamma = FALSE;
 
@@ -1344,37 +1345,35 @@ Rboolean SVGDeviceDriver(NewDevDesc *dd, char *filename, char *bg, char *fg,
     return TRUE;
 }
 
-static  GEDevDesc *RSvgDevice(char **file, char **bg, char **fg,
+static  pGEDevDesc RSvgDevice(char **file, char **bg, char **fg,
                               double *width, double *height, int *debug,
                               int *xmlHeader, char **title, int *toolTipMode,
 			      int *tipFontSize, double *tipOpacity,
 			      int *onefile, int *useStyleAttributes)
 {
-    GEDevDesc *dd;
-    NewDevDesc *dev;
+    pGEDevDesc dd;
+    pDevDesc dev;
 
     if (debug[0] == NA_LOGICAL) debug = FALSE;
 
+    R_GE_checkVersionOrDie(R_GE_version);
     R_CheckDeviceAvailable();
     BEGIN_SUSPEND_INTERRUPTS {
         if (!(dev = Calloc(1, NewDevDesc)))
             error("unable to allocate memory for NewDevDesc (small)");
-        /* Do this for early redraw attempts */
-        dev->displayList = R_NilValue;
-
         if (!SVGDeviceDriver(dev, file[0], bg[0], fg[0], width[0], height[0], debug[0],
 			     xmlHeader[0], title[0], toolTipMode[0], tipFontSize[0],
 			     tipOpacity[0], onefile[0], useStyleAttributes[0])) {
             Free(dev);
             error("unable to start device SVG");
         }
-        gsetVar(install(".Device"), mkString("devSVG"), R_NilValue);
         dd = GEcreateDevDesc(dev);
 
 #if R_VERSION < R_Version(2,7,0)
+        gsetVar(install(".Device"), mkString("devSVG"), R_NilValue);
 	Rf_addDevice((DevDesc*) dd);
 #else
-        GEaddDevice(dd);
+        GEaddDevice2(dd, "devSVG");
 #endif
         GEinitDisplayList(dd);
     } END_SUSPEND_INTERRUPTS;
